// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


/*
   Establece un manejador externo para controlar el teclado, recoge los parametros de la operacion
   por este y envia la peticion al servidor que devolvera el resultado de aplicar la operacion 
   sobre los parametros recogidos.

   El cliente de esta aplicacion es: server.p => Transporte: comm::Transport.
*/
#include <iostream>

#include <string.h>

#include <anna/core/core.hpp>
#include <anna/comm/comm.hpp>

#include <anna/test/Menu.hpp>
#include <anna/test/Request.hpp>
#include <anna/test/Response.hpp>

class MyCommunicator : public Communicator {
public:   
   MyCommunicator () : Communicator () {;}
   
private:
   test::Response a_response;   
   test::Request a_request;
   
   void eventReceiveMessage (ClientSocket&, const Message&) throw (RuntimeException);
   void eventUser (const char* id, const void* context) throw ();   
};

class KClient : public anna::comm::Application {
public:
   KClient ();
      
   Server* getServer () const throw () { return a_server; }
   const test::Menu& getMenu () const throw () { return a_menu; }
   
private:
   MyCommunicator a_communicator;
   test::Menu a_menu;
   Server* a_server;

   void initialize () throw (RuntimeException);
   void run () throw (RuntimeException);    
};

using namespace std;
using namespace test;

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   KClient app;
   
   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      Logger::setLevel (Logger::Debug); 
      Logger::initialize ("kclient", new TraceWriter ("file.trace", 4096000));
 
      app.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }
   
   return 0;
}

KClient::KClient () : 
   Application ("kclient", "KClient", "1.0.0"),
   a_menu (&a_communicator)
{
   CommandLine& commandLine (CommandLine::instantiate ());
      
   commandLine.add ("p", CommandLine::Argument::Mandatory, "Puerto en el que el servidor atiende respuestas.");
   commandLine.add ("a", CommandLine::Argument::Optional, "Direccion IP Puerto en el que el servidor atiende respuestas.");
   commandLine.add ("host", CommandLine::Argument::Optional, "Nombre del host en el que atiende el servidor");
   commandLine.add ("auto", CommandLine::Argument::Optional, "Autoreconexion", false);
   commandLine.add ("trace", CommandLine::Argument::Optional, "Nivel de trazas (debug,warning, error,...)");

   activateGeneralPublicLicense ();
}

void KClient::initialize () 
   throw (RuntimeException)
{
   LOGMETHOD (TraceMethod tm ("MyCommunicator", "initialize", ANNA_FILE_LOCATION));

   CommandLine& cl (CommandLine::instantiate ());    

   Network& network = Network::instantiate ();

  if (cl.exists ("trace"))
      Logger::setLevel (Logger::asLevel (cl.getValue ("trace")));

   const int p = cl.getIntegerValue ("p");

   if (cl.exists ("a"))
      a_server = network.createServer (cl.getValue ("a"), p, cl.exists ("auto"));
   else if (cl.exists ("host")) {
      a_server = network.resolveServer (cl.getValue ("host"), p, cl.exists ("auto"));
   }
   else 
      throw RuntimeException ("Hay que indicar el argumento 'a' o 'host'", ANNA_FILE_LOCATION);

   a_communicator.attach (&a_menu);
}

void KClient::run ()
   throw (RuntimeException)
{   
   LOGMETHOD (TraceMethod tm ("MyCommunicator", "run", ANNA_FILE_LOCATION));

   a_menu.paint ();
   a_communicator.accept ();
}

void MyCommunicator::eventReceiveMessage (ClientSocket&, const Message& message)
   throw (RuntimeException)
{
   LOGMETHOD (TraceMethod tm ("MyCommunicator", "eventReceiveMessage", ANNA_FILE_LOCATION));

   a_response.decode (message.getBody ());

   const Millisecond responseTime = (Millisecond)(anna::functions::microsecond()) - a_response.initTime;
   
   cout << endl << "ResponseTime: " << responseTime << " us" << endl;
   cout << "Resultado de la peticion: " << a_response.x << (char) a_response.op << a_response.y  << " = " << a_response.result << endl << endl;
   
   static_cast <KClient&> (anna::comm::functions::getApp ()).getMenu ().paint ();
}

//-----------------------------------------------------------------------------------------
// Cuando el Menu tiene disponibles todos los datos necesarios para la peticin se lo
// notifica al comunicador mediante un evento de usuario.
//-----------------------------------------------------------------------------------------
void MyCommunicator::eventUser (const char* id, const void* context) 
   throw ()
{
   LOGMETHOD (TraceMethod tm ("MyCommunicator", "eventUser", ANNA_FILE_LOCATION));

   if (anna_strcmp (id, Menu::EventData) == 0) {
      const Menu::Data* data (reinterpret_cast <const Menu::Data*>  (context));
      
      a_request.op = data->a_operation;
      a_request.x = data->a_op1;
      a_request.y = data->a_op2;
      a_request.initTime = anna::functions::microsecond ();

      Server* server = static_cast <KClient&> (anna::comm::functions::getApp ()).getServer ();
      
      try {
         server->send (a_request);
      }
      catch (RuntimeException& ex) {
         ex.trace ();
      }
   } 
}
