// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <iostream>

#include <functional>

#include <anna/core/core.hpp>

#include <anna/app/Application.hpp>

class Test : public app::Application {
public:
   Test ();
   
   void initialize () throw (RuntimeException);
   
private:
   void  run () throw (RuntimeException);      
};

using namespace std;

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   Test test;

   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      if (commandLine.exists ("t"))
         Logger::setLevel (Logger::asLevel (commandLine.getValue ("t"))); 

      Logger::initialize ("testfunctions", new TraceWriter ("file.trace", 2048000));
 
      test.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }
   
   return 0;
}

Test::Test () : 
   app::Application ("testfunctions", "Checking trace system", "1.0") 
{
   CommandLine& commandLine (CommandLine::instantiate ());
      
   commandLine.add ("s", CommandLine::Argument::Mandatory, "String");
   commandLine.add ("n", CommandLine::Argument::Mandatory, "Number of loops");
   commandLine.add ("t", CommandLine::Argument::Optional, "Initial trace level");
}

void Test::initialize () 
   throw (RuntimeException)
{
   CommandLine& commandLine (CommandLine::instantiate ());

   if (commandLine.exists ("t"))
      Logger::setLevel (Logger::asLevel (commandLine.getValue ("t"))); 
}

void Test::run () 
   throw (RuntimeException)
{
   CommandLine& commandLine (CommandLine::instantiate ());

   cout << "Second: " << functions::second () << endl;
   cout << "MilliSecond: " << functions::millisecond () << endl << endl;

   string str (commandLine.getValue ("s"));
   int n (commandLine.getIntegerValue ("n"));
   
   Logger::Level level = Logger::getLevel ();

   for (int i = 0; i < n; i ++) {
      if ((i % 2) == 0)
         Logger::write (level, "string", str.c_str (), ANNA_FILE_LOCATION);
      else
         Logger::write (level, "level", (int) level, ANNA_FILE_LOCATION);
         
      level = Logger::Level ((int) level + 1);
      
      if (level >= Logger::Local0)
         level = Logger::Emergency;
      else if (level > Logger::Debug)
         level = Logger::Local0;
   }
}

