#!/bin/bash

#############
# VARIABLES #
#############

# General
EXE_BN=ADML

#############
# FUNCTIONS #
#############

_exit () {
  echo
  echo $1
  echo
  exit 1
}

# $1: process name
createRunScript () {

  cat << EOF > run.sh
#!/bin/bash
cd \`dirname \$0\`
STARTED=\`pgrep $1$ 2>/dev/null\`
[ \$? -eq 0 ] && { echo "Already started!"; echo "\$STARTED" ; exit 1 ; }
0> launcher.trace
rm -f counters/* test-reports/*
# Execution line:
export LD_LIBRARY_PATH=\$PWD/dynlibs
./$@ --services services.xml &
echo \$! > .pid
EOF

  chmod a+x run.sh
}

# $1: input option (c, s, b, d)
# Retuns the type of launcher logical name for the input option
get_tol () {
  [ "$1" = "c" ] && echo client
  [ "$1" = "s" ] && echo server
  [ "$1" = "b" ] && echo balancer
  [ "$1" = "d" ] && echo dummy
}

#############
# EXECUTION #
#############
cd `dirname $0`
SCR_DIR=`pwd`
echo
echo
echo "Configure a (c)lient, (s)erver, (b)alancer or (d)ummy [c]:"
read option
[ "$option" = "" ] && option=c
tol=$(get_tol $option)
[ "$tol" = "" ] && _exit "Option '$option' not implemented !!"

# Services
ln -sf services/${tol}.xml services.xml

# Tracing
DEBUG="--cntDir counters --tmDir test-reports"
echo
echo "Enable debug traces ? (y/n) [n]:"
read enable
[ "$enable" = "" ] && enable=n
[ "$enable" = "y" ] && DEBUG="$DEBUG --trace debug"

# Kindness
KINDNESS=
echo
echo "Strict xml for decoded messages ? (y/n) [y]:"
read strict
[ "$strict" = "" ] && strict=y
[ "$strict" = "n" ] && KINDNESS="--ignoreFlags"
if [ "$option" = "s" ]
then
  echo
  echo "Ignore errors ? (y/n) [n]:"
  read i_errors
  [ "$i_errors" = "" ] && i_errors=n
  [ "$i_errors" = "y" ] && KINDNESS="$KINDNESS --ignoreErrors"
fi

# Run script:
EXE_LINK=`basename $PWD`
[ "$EXE_BN" != "$EXE_LINK" ] && ln -sf $EXE_BN $EXE_LINK

case $tol in

  client)
    ENTITY=$LOCAL_STANDARD_ENDPOINT
    createRunScript $EXE_LINK $KINDNESS $DEBUG &
    ;;

  server)
    SERVER=$LOCAL_STANDARD_ENDPOINT
    createRunScript $EXE_LINK $KINDNESS $DEBUG &
    ;;

  balancer)
    SERVER=$LOCAL_STANDARD_ENDPOINT
    ENTITY=$EXAMPLE_ENTITY_4_BALANCER
    createRunScript $EXE_LINK $KINDNESS $DEBUG &
    ;;

  dummy)
    createRunScript $EXE_LINK $KINDNESS $DEBUG &
    ;;

esac

echo
echo "Created 'run.sh' script !"
echo

