// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_LiteTransport_hpp
#define anna_comm_LiteTransport_hpp

#include <anna/comm/Transport.hpp>
#include <anna/comm/TransportFactory.hpp>
#include <anna/comm/TransportFactoryImpl.hpp>

namespace anna {

namespace comm {

/**
  Protocol oriented to internal network messages interchange, supposing no errors in transmision and short-sized
  messages which allows length representation in only two bytes. No synchronization mechanisms are provided, then
  a single error could cause problems in our process.

  @warning Encode mode does not provide enough information to allow re-synchronization in case of error

  \see Transport.
*/
class LiteTransport : public comm::Transport {
public:
  static const int headerSize = sizeof(short int);  /**< Tamao de la cabezera del mensaje */

  /**
    Destructor.
  */
  virtual ~LiteTransport();

  /**
    Returns the class name literal.
    @return class name literal.
  */
  static const char* className() throw() { return "anna::comm::LiteTransport"; }

  /**
    Returns the transport factory associated to this class
    @return transport factory associated to this class
  */
  static TransportFactory& getFactory() throw() { return st_transportFactory; }

private:
  static TransportFactoryImpl <LiteTransport> st_transportFactory;

  LiteTransport();

  int calculeSize(const DataBlock&) throw(RuntimeException);
  const Message* decode(const DataBlock&) throw(RuntimeException);
  const DataBlock& code(Message&) throw(RuntimeException);

  friend class anna::Allocator <LiteTransport>;
};

}
}
#endif

