// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_EncodedData_hpp
#define anna_core_util_EncodedData_hpp

#include <openssl/des.h>

#include <anna/core//DataBlock.hpp>

namespace anna {

namespace xml {
class Node;
}

class Encoder;

/**
   Interface generico para guardar el resultado de una codificacion. Ver anna::util::Encoder.
*/
class EncodedData {
public:
  /**
     Constructor.
  */
  EncodedData() : a_value(true), a_realSize(0) {;}

  /**
     Inicializa el contenido de esta clase a partir de la informacion contenida en el
     documento XML recibido.

     \param document Documento XML que contiene la informacion con la inicializar esta clase.
     Deberia haber sido obtenido mediate #asXML.
  */
  void initialize(const xml::Node* document) throw(RuntimeException);

  /**
     Devuelve un documento XML con la informacion necesaria para reproducir el contenido de esta
     instancia.
     \param parent Nodo XML del que dependende la informacion.
     \return Nodo XML que en el que podemos aplicar informacion.
  */
  xml::Node* asXML(xml::Node* parent) const throw(RuntimeException);

private:
  DES_key_schedule a_skey [3];
  DES_cblock a_iv;
  DataBlock a_value;
  int a_realSize;

  EncodedData(const EncodedData&);

  static const std::string& asBCD(const DataBlock&, std::string&) throw();
  static int fromBCD(const std::string&, DataBlock&) throw();

  friend class Encoder;
};


}

#endif
