// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/DataBlock.hpp>
#include <anna/config/defines.hpp>
#include <anna/core/functions.hpp>

#include <anna/comm/DatagramSocket.hpp>

using namespace std;
using namespace anna;

comm::DatagramSocket::DatagramSocket(const comm::DatagramSocket::Mode mode, const comm::INetAddress& address, comm::TransportFactory* transportFactory) :
      comm::ClientSocket(transportFactory, comm::Socket::Domain::Inet, comm::Socket::Type::Datagram),
      a_mode(mode) {
   if (mode == ReadOnly)
      a_localAccessPoint = address;
   else
      a_remoteAccessPoint = address;
}

void comm::DatagramSocket::connect()
throw(RuntimeException) {
   Guard guard(*this, "comm::DatagramSocket::connect");
   anna_socket_assert(isConnected() == true, "Already connected");

   if (Socket::isOpened() == false) {
      Socket::open();
      getSocketOptions();
      setBlockingMode(false);
   }

   // Hace el bind a la localAddress.
   if (a_mode == ReadOnly && Socket::isBound() == false)
      Socket::bind();

   activate(Status::Connected);
   LOGDEBUG(
      string msg("comm::DatagramSocket::connect | ");
      msg += asString();
      Logger::debug(msg, ANNA_FILE_LOCATION);
   );
}

void comm::DatagramSocket::do_write(const DataBlock& message)
throw(RuntimeException) {
   sockaddr* s(NULL);
   int len(0);
   a_remoteAccessPoint.translate(*this, s, len);
   int r;
   anna_signal_shield(r, sendto(Socket::a_fd, message.getData(), message.getSize(), 0, s, len));

   if (r < 0) {
      const int xerrno = errno;
      string msg(asString());
      msg += " | Cannot be sent";
      throw RuntimeException(msg, xerrno, ANNA_FILE_LOCATION);
   }

   LOGDEBUG(
      string msg("comm::DatagramSocket::do_write | ");
      msg += asString();
      msg += functions::asText(" | Sent: ", message);
      Logger::write(Logger::Debug, msg, ANNA_FILE_LOCATION)
   );
}

int comm::DatagramSocket::do_read(const char* data, const int maxSize)
throw(RuntimeException) {
   int result;
   anna_signal_shield(result, recvfrom(Socket::a_fd, (void*) data, maxSize, 0, (sockaddr*) NULL, NULL));

   if (result < 0) {
      const int xerrno = errno;
      string msg(asString());
      msg += " | Cannot be received";
      throw RuntimeException(msg, xerrno, ANNA_FILE_LOCATION);
   }

   LOGDEBUG(
      string msg("comm::DatagramSocket::do_read | ");
      msg += asString();
      msg += functions::asText(" | N-bytes", result);
      msg += functions::asText(" | Received: ", DataBlock(data, result, false));
      Logger::write(Logger::Debug, msg, ANNA_FILE_LOCATION)
   );
   return result;
}


