// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/comm/internal/Poll.hpp>

//-----------------------------------------------------------------------------------
//  RECORDAR QUE ESTA CLASE SOLO SE USA EN LA VERSION SINGLE-THREAD DEL
//  comm::Communicator
//-----------------------------------------------------------------------------------

using namespace anna;

void comm::Poll::setTimeout(const Millisecond &timeout)
throw() {
  if((a_timeout.tv_sec = timeout / 1000) == 0)
    a_timeout.tv_usec = timeout * 1000;
  else
    a_timeout.tv_usec = 0;

  a_ptrTimeout = (a_timeout.tv_sec != 0 || a_timeout.tv_usec != 0) ? &a_timeout : NULL;
}

//--------------------------------------------------------------------------------------------
// Si a_ptrTimeout != NULL => se devuelve el tiempo que ha sobrado de la espera maxima
// establecida (en Solaris este puntero no se modifica para nada)
//
// Para las dos implementacion el NULL => espera hasta que llegue un mensaje.
//--------------------------------------------------------------------------------------------
void comm::Poll::waitMessage()
throw(RuntimeException) {
  if(a_ptrTimeout == NULL)
    a_pollr = select(a_maxfd, (fd_set*) anna_memcpy(&a_fdset, &a_fdmask, sizeof(fd_set)), NULL);
  else {
#ifdef __linux__
    timeval timeout(a_timeout);
    a_pollr = select(a_maxfd, (fd_set*) anna_memcpy(&a_fdset, &a_fdmask, sizeof(fd_set)), &timeout);
#else
    a_pollr = select(a_maxfd, (fd_set*) anna_memcpy(&a_fdset, &a_fdmask, sizeof(fd_set)), a_ptrTimeout);
#endif
  }
}

//-----------------------------------------------------------------------------------------
// Se ha detectado una condicion de error. Se recibe trafico por fd = { 16, 19} y el
// tramiento de 16 => cierre de 19 mediante comm::Communicator::attach => el a_ifd no
// llegaba nunca a tratar al 19, porque el a_maxfd ha sido modificado con 16, pero el a_pollr
// sigue indicado que queda un fd por tratar. ver Poll::erase
//
// (1) Una vez que ha tratado detectado la activada desactiva el indicador
//-----------------------------------------------------------------------------------------
int comm::Poll::fetch()
throw() {
  int result = -1;

  while(a_pollr > 0) {
    if(FD_ISSET(a_ifd, &a_fdset)) {
      FD_CLR(a_ifd, &a_fdset);                                           // (1)
      result = a_ifd ++;

      if(a_ifd > a_maxfd)
        a_ifd = a_minfd;

      a_pollr --;
      break;
    }

    if(++ a_ifd > a_maxfd)
      a_ifd = a_minfd;
  }

  return result;
}

void comm::Poll::insert(const int fd)
throw() {
  FD_SET(fd, &a_fdmask);

  if(a_maxfd < fd) a_maxfd = fd;

  if(fd < a_minfd)
    a_ifd = a_minfd = fd;
}

//---------------------------------------------------------------------------------------
// Hay que comprobar que el 'fd' que vamos a eliminar de la lista no este en la lista
// de pendientes de tratar.
//---------------------------------------------------------------------------------------
void comm::Poll::erase(const int fd)
throw() {
  FD_CLR(fd, &a_fdmask);

  if(FD_ISSET(fd, &a_fdset) && a_pollr > 0) {                                // (1)
    FD_CLR(fd, &a_fdset);
    a_pollr --;
  }

  if(fd == a_maxfd || fd == a_minfd) {
    const int max = a_maxfd;
    a_maxfd = 0;
    a_minfd = INT_MAX;

    for(int ifd = 0; ifd <= max; ifd ++) {
      if(FD_ISSET(ifd, &a_fdmask)) {
        if(a_maxfd < ifd) a_maxfd = ifd;

        if(ifd < a_minfd) a_minfd = ifd;
      }
    }

    a_ifd = a_minfd;
  }
}

