// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/core/util/RegularExpression.hpp>

#include <anna/core/util/Tokenizer.hpp>
#include <anna/core/functions.hpp>


//using namespace anna;

// private

void anna::RegularExpression::freeRegex() throw() {
  if(a_compiled) {
    regfree(&a_preg);
    a_compiled = false;
  }
}

void anna::RegularExpression::compile() throw(anna::RuntimeException) {
  if(a_compiled) return;

  int ret;

  if((ret = regcomp(&a_preg, a_pattern.c_str(), REG_EXTENDED)) != 0) {
    char err[256];
    std::string msg("anna::RegularExpression::setPattern | ");
    msg += " | Pattern: ";
    msg += a_pattern;
    msg += " | ";

    if(regerror(ret, &a_preg, err, sizeof(err)))
      msg += err;
    else
      msg += "Invalid pattern provided";

    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  a_compiled = true;
}

// public

//------------------------------------------------------------------------------
//---------------------------------------------- RegularExpression::setPattern()
//------------------------------------------------------------------------------
void anna::RegularExpression::setPattern(const std::string & pattern) throw() {
  if(pattern == a_pattern) return;

  freeRegex();
  a_pattern = pattern;
}


//------------------------------------------------------------------------------
//-------------------------------------------------- RegularExpression::isLike()
//------------------------------------------------------------------------------
bool anna::RegularExpression::isLike(const std::string & value) throw() {
  compile();
  const bool result = (regexec(&a_preg, value.c_str(), 0, NULL, 0) == 0) ? true : false;
  return result;
}


//------------------------------------------------------------------------------
//------------------------------------------------ RegularExpression::operator==
//------------------------------------------------------------------------------
bool anna::RegularExpression::operator == (const RegularExpression & re) const {
  return (getPattern() == re.getPattern());
}


//------------------------------------------------------------------------------
//------------------------------------------------- RegularExpression::operator<
//------------------------------------------------------------------------------
bool anna::RegularExpression::operator < (const RegularExpression & re) const {
  return (getPattern() < re.getPattern());
}

