// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <arpa/inet.h>

#include <zlib.h>

#include <anna/core/util/ZBlock.hpp>

using namespace std;
using namespace anna;

/*
 * El bloque de datos retornoado contiene:
 * <int que indica el tamao original del buffer> + buffer comprimido.
 */
const DataBlock& ZBlock::compress(const DataBlock& data, const Mode::_v mode)
throw(RuntimeException) {
  if(&data == this)
    throw RuntimeException("ZBlock::compress | Source and target object cannot be the same", ANNA_FILE_LOCATION);

  int originalSize;

  /*
   * Si la fuente de datos est vaca no hay que hacer nada
   */
  if((originalSize = data.getSize()) == 0) {
    DataBlock::clear();
    return *this;
  }

  /*
   * Segn la documentacin el buffer destino debe ser, como mnimo un 0.1% mayor que el original + 12 bytes
   * Lo incrementamos en un 1% para simplificar
   */
  uLong maxSize = sizeof(int) + 12 + originalSize + (originalSize * 101) / 100;
  DataBlock::allocate(maxSize);
  char* buffer = (char*) DataBlock::getData();
  int aux = htonl(originalSize);
  char* w((char*) &aux);
  /* Establece el tamao original del buffer */
  buffer [0] = *w;
  buffer [1] = *(w + 1);
  buffer [2] = *(w + 2);
  buffer [3] = *(w + 3);
  int rr;

  if(mode == Mode::Default)
    rr = ::compress((Bytef*) buffer + sizeof(int), &maxSize, (Bytef*) data.getData(), originalSize);
  else
    rr = ::compress2((Bytef*) buffer + sizeof(int), &maxSize, (Bytef*) data.getData(), originalSize, mode);

  if(rr != Z_OK) {
    string msg("ZBlock::compress | ");
    msg += zError(rr);
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  DataBlock::setSize((int) maxSize + sizeof(int));
  return *this;
}

const DataBlock& ZBlock::uncompress(const DataBlock& zdata)
throw(RuntimeException) {
  if(&zdata == this)
    throw RuntimeException("ZBlock::uncompress | Source and target object cannot be the same", ANNA_FILE_LOCATION);

  /*
   * Si la fuente de datos est vaca no hay que hacer nada
   */
  if(zdata.getSize() < sizeof(int)) {
    DataBlock::clear();
    return *this;
  }

  char* buffer = (char*) zdata.getData();
  int aux;
  uLong size;
  char* w((char*) &aux);
  *w  = *buffer;
  *(w + 1) = *(buffer + 1);
  *(w + 2) = *(buffer + 2);
  *(w + 3) = *(buffer + 3);
  size = ntohl(aux);
  DataBlock::allocate((int) size);
  int rr = ::uncompress((Bytef*) DataBlock::getData(), &size, (const Bytef*) zdata.getData() + sizeof(int), zdata.getSize() - sizeof(int));

  if(rr != Z_OK) {
    string msg("ZBlock::uncompress | ");
    msg += zError(rr);
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  DataBlock::setSize((int) size);
  return *this;
}
